use std::{
    cell::RefCell,
    collections::{HashMap, HashSet},
    marker::PhantomData,
};

use avian2d::prelude::*;
use bevy::prelude::*;
use coord_2d::{Coord, Size};
use shadowcast::{vision_distance, Context, InputGrid};

use crate::{
    core::{GlobalTransformExt, Player, PointLike},
    log::Log,
    map::{Map, MapPlugin},
};

#[derive(Component, Clone, Copy, Debug, Default, Reflect)]
#[reflect(Component)]
pub struct DontLogWhenVisible;

#[derive(Component, Clone, Debug, Default, Deref, DerefMut, Reflect)]
#[reflect(Component)]
pub struct RevealedTiles(pub Vec<bool>);

#[derive(Component, Clone, Debug, Eq, PartialEq)]
pub struct Viewshed {
    pub range: u32,
    pub visible_points: HashSet<(i32, i32)>,
}

impl Default for Viewshed {
    fn default() -> Self {
        Self {
            range: 15,
            visible_points: HashSet::new(),
        }
    }
}

impl Viewshed {
    pub fn is_point_visible(&self, point: &dyn PointLike) -> bool {
        self.visible_points.contains(&point.into())
    }

    fn update(
        &mut self,
        commands: &mut Commands,
        viewer_entity: &Entity,
        visible_entities: &mut VisibleEntities,
        start: &Vec2,
        spatial_query: &SpatialQuery,
        visible_query: &Query<(&Visible, &Collider, &GlobalTransform)>,
        cache: &mut RefCell<HashMap<(i32, i32), (u8, HashSet<Entity>)>>,
    ) {
        // println!("Start");
        let mut boxes = vec![];
        let shape = Collider::rectangle(self.range as f32, self.range as f32);
        spatial_query.shape_intersections_callback(&shape, *start, 0., default(), |entity| {
            if let Ok((_, collider, transform)) = visible_query.get(entity) {
                boxes.push(collider.aabb(transform.translation().truncate(), 0.));
            }
            true
        });
        let mut context: Context<u8> = Context::default();
        let vision_distance = vision_distance::Circle::new(self.range);
        let shape = Collider::rectangle(0.49, 0.49);
        let size = (
            (start.x.abs() + self.range as f32 * 2.) as u32,
            (start.y.abs() + self.range as f32 * 2.) as u32,
        );
        let visibility_grid = VisibilityGrid(
            size,
            RefCell::new(Box::new(|coord: Coord| {
                let shape_pos = Vec2::new(coord.x as f32 + 0.5, coord.y as f32 + 0.5);
                // println!("Checking {coord:?}: {shape_pos:?}");
                if start.distance(&shape_pos) > self.range as f32 {
                    // println!("Out of range");
                    return u8::MAX;
                }
                let mut coord_entities = HashSet::new();
                let mut to_insert = None;
                let opacity = if let Some((opacity, entities)) =
                    cache.borrow().get(&(coord.x, coord.y))
                {
                    // println!("Cache hit: {opacity:?}: {entities:?}");
                    coord_entities = entities.clone();
                    *opacity
                } else {
                    let aabb = shape.aabb(shape_pos, 0.);
                    if boxes.iter().any(|b| b.intersects(&aabb)) {
                        // println!("Hit");
                        let mut opacity = 0;
                        spatial_query.shape_intersections_callback(
                            &shape,
                            shape_pos,
                            0.,
                            default(),
                            |entity| {
                                if let Ok((visible, _, _)) = visible_query.get(entity) {
                                    // println!(
                                    // "{entity:?}: {visible:?}: {:?}",
                                    // transform.translation().truncate()
                                    // );
                                    coord_entities.insert(entity);
                                    opacity = opacity.max(**visible);
                                }
                                true
                            },
                        );
                        to_insert = Some(((coord.x, coord.y), (opacity, coord_entities.clone())));
                        // println!("New opacity: {opacity}");
                        opacity
                    } else {
                        // println!("No hits, 0");
                        to_insert = Some(((coord.x, coord.y), default()));
                        0
                    }
                };
                if let Some((k, v)) = to_insert {
                    cache.borrow_mut().insert(k, v);
                }
                if coord_entities.contains(viewer_entity) {
                    let mut coord_entities = coord_entities.clone();
                    coord_entities.retain(|e| e != viewer_entity);
                    let opacity = coord_entities
                        .iter()
                        .map(|v| visible_query.get(*v).unwrap().0 .0)
                        .max()
                        .unwrap_or(0);
                    // println!("Viewer hit, removing viewer opacity: {opacity}");
                    opacity
                } else {
                    opacity
                }
            })),
        );
        let mut new_visible = HashSet::new();
        let mut new_visible_entities = HashSet::new();
        // println!("Start: {viewer_entity}");
        context.for_each_visible(
            Coord::new(start.x_i32(), start.y_i32()),
            &visibility_grid,
            &size,
            vision_distance,
            u8::MAX,
            |coord, _directions, _visibility| {
                new_visible.insert((coord.x, coord.y));
                if let Some((_, entities)) = cache.borrow().get(&(coord.x, coord.y)) {
                    for e in entities {
                        new_visible_entities.insert(*e);
                    }
                }
            },
        );
        if self.visible_points != new_visible {
            self.visible_points = new_visible;
        }
        if new_visible_entities != **visible_entities {
            for lost in visible_entities.difference(&new_visible_entities) {
                commands.trigger_targets(VisibilityChanged::Lost(*lost), *viewer_entity);
            }
            for gained in new_visible_entities.difference(visible_entities) {
                // println!("transition: {:?} gains {:?}", viewer_entity,
                // gained);
                commands.trigger_targets(VisibilityChanged::Gained(*gained), *viewer_entity);
            }
            **visible_entities = new_visible_entities;
        }
    }
}

#[derive(Component, Clone, Copy, Debug, Deref, DerefMut, Eq, PartialEq, Reflect)]
#[reflect(Component)]
pub struct Visible(pub u8);

impl Default for Visible {
    fn default() -> Self {
        Self::opaque()
    }
}

impl Visible {
    pub fn transparent() -> Self {
        Self(0)
    }

    pub fn opaque() -> Self {
        Self(u8::MAX)
    }
}

#[derive(Component, Clone, Debug, Default, Deref, DerefMut)]
pub struct VisibleEntities(HashSet<Entity>);

#[derive(Bundle, Default)]
pub struct ViewshedBundle {
    pub viewshed: Viewshed,
    pub visible_entities: VisibleEntities,
}

impl ViewshedBundle {
    pub fn new(range: u32) -> Self {
        Self {
            viewshed: Viewshed { range, ..default() },
            ..default()
        }
    }
}

#[derive(Event)]
pub enum VisibilityChanged {
    Gained(Entity),
    Lost(Entity),
}

impl PointLike for Coord {
    fn x(&self) -> f32 {
        self.x as f32
    }

    fn y(&self) -> f32 {
        self.y as f32
    }
}

fn add_visibility_indices<MapData: 'static + Clone + Default + Send + Sync>(
    mut commands: Commands,
    map_config: Res<MapPlugin<MapData>>,
    query: Query<(Entity, &Map<MapData>), (Added<Map<MapData>>, Without<RevealedTiles>)>,
) {
    for (entity, map) in &query {
        let count = map.width * map.height;
        commands
            .entity(entity)
            .insert(RevealedTiles(vec![map_config.start_revealed; count]));
    }
}

fn viewshed_removed(
    mut commands: Commands,
    mut query: RemovedComponents<Viewshed>,
    visible_entities: Query<&VisibleEntities>,
) {
    for entity in query.read() {
        if let Ok(visible) = visible_entities.get(entity) {
            for e in visible.iter() {
                commands.trigger_targets(VisibilityChanged::Lost(*e), entity);
            }
        }
    }
}

pub struct VisibilityGrid<F>(pub (u32, u32), pub RefCell<Box<F>>);

impl<F> InputGrid for VisibilityGrid<F>
where
    F: FnMut(Coord) -> u8,
{
    type Grid = (u32, u32);

    type Opacity = u8;

    fn size(&self, grid: &Self::Grid) -> Size {
        Size::new(grid.0, grid.1)
    }

    fn get_opacity(&self, _grid: &Self::Grid, coord: Coord) -> Self::Opacity {
        (self.1.borrow_mut())(coord)
    }
}

fn update_viewshed(
    mut commands: Commands,
    visible: Query<(&Visible, &Collider, &GlobalTransform)>,
    mut viewers: Query<(
        Entity,
        &mut Viewshed,
        &mut VisibleEntities,
        &GlobalTransform,
    )>,
    spatial_query: SpatialQuery,
) {
    let mut cache = default();
    for (viewer_entity, mut viewshed, mut visible_entities, viewer_transform) in &mut viewers {
        viewshed.update(
            &mut commands,
            &viewer_entity,
            &mut visible_entities,
            &viewer_transform.translation().truncate(),
            &spatial_query,
            &visible,
            &mut cache,
        );
    }
}

fn remove_visible(
    mut commands: Commands,
    mut removed: RemovedComponents<Visible>,
    mut viewers: Query<(
        Entity,
        &mut Viewshed,
        &mut VisibleEntities,
        &GlobalTransform,
    )>,
    spatial_query: SpatialQuery,
    visible: Query<(&Visible, &Collider, &GlobalTransform)>,
) {
    if !removed.is_empty() {
        let mut cache = default();
        for removed in removed.read() {
            for (viewer_entity, mut viewshed, mut visible_entities, start) in &mut viewers {
                if !visible_entities.contains(&removed) {
                    continue;
                }
                viewshed.update(
                    &mut commands,
                    &viewer_entity,
                    &mut visible_entities,
                    &start.translation().truncate(),
                    &spatial_query,
                    &visible,
                    &mut cache,
                );
            }
        }
    }
}

fn update_revealed_tiles<D: 'static + Clone + Default + Send + Sync>(
    mut map: Query<(&Map<D>, &mut RevealedTiles)>,
    viewers: Query<&Viewshed, (With<Player>, Changed<Viewshed>)>,
) {
    for viewshed in &viewers {
        for (map, mut revealed_tiles) in &mut map {
            for v in viewshed.visible_points.iter() {
                let idx = v.to_index(map.width);
                if idx >= revealed_tiles.len() {
                    continue;
                }
                revealed_tiles[idx] = true;
            }
        }
    }
}

fn log_visible(
    trigger: Trigger<VisibilityChanged>,
    viewers: Query<(&GlobalTransform, Option<&Collider>), (With<Player>, With<Viewshed>)>,
    visible: Query<(&Name, &GlobalTransform, Option<&Collider>), Without<DontLogWhenVisible>>,
    mut log: Query<&mut Log>,
) {
    if let VisibilityChanged::Gained(entity) = trigger.event() {
        if *entity == trigger.entity() {
            return;
        }
        if let Ok((viewer_transform, viewer_collider)) = viewers.get(trigger.entity()) {
            if let Ok((name, viewed_transform, viewed_collider)) = visible.get(*entity) {
                // println!("Gain {name}: {entity}");
                let location = if let (Some(viewer_collider), Some(viewed_collider)) =
                    (viewer_collider, viewed_collider)
                {
                    viewer_transform.collider_direction_and_distance(
                        viewer_collider,
                        viewed_transform,
                        viewed_collider,
                    )
                } else {
                    let yaw = viewer_transform.yaw();
                    viewer_transform.direction_and_distance(viewed_transform, Some(yaw))
                };
                if let Ok(mut log) = log.get_single_mut() {
                    log.push(format!("{}: {location}", name));
                }
            }
        }
    }
}

pub struct VisibilityPlugin<MapData: 'static + Clone + Default + Send + Sync>(PhantomData<MapData>);

impl<MapData: 'static + Clone + Default + Send + Sync> Default for VisibilityPlugin<MapData> {
    fn default() -> Self {
        Self(Default::default())
    }
}

impl<MapData: 'static + Clone + Default + Send + Sync> Plugin for VisibilityPlugin<MapData> {
    fn build(&self, app: &mut App) {
        app.add_systems(
            FixedPreUpdate,
            (
                add_visibility_indices::<MapData>,
                update_viewshed,
                update_revealed_tiles::<MapData>,
            ),
        )
        .add_systems(FixedPostUpdate, (viewshed_removed, remove_visible))
        .observe(log_visible);
    }
}
